﻿using System.Collections.Generic;

namespace Sharing
{
	static class Algorithm
	{
		private static TermGraph tg;
		
		private static void BuildEquivalenceClass(Node c)
		{
			c.Queue.Enqueue(c);
			c.Canonic = c;
			c.Building = true;
			
			while (c.Queue.Count > 0)
			{
				Node n = c.Queue.Dequeue();
				foreach (Node m in n.Parents)
				{
					if (m.Canonic == null)
					{
						BuildEquivalenceClass(m);
						tg.Print();
					} else if (m.Canonic.Building)
					{
						throw new ExecutionException("Cycle Found");
					}
				}
				foreach (UndirectedEdge edge in n.UndirectedEdges)
				{
					Node m = edge.FindTarget(n);
					if (m.Canonic == null)
					{
						EnqueueAndPropagate(m, c);
						tg.Print();
					} else if (m.Canonic != c)
					{
						throw new ExecutionException("Cycle Found");
					}
				}
			}
			c.Building = false;
			tg.Print();
		}

		private static void EnqueueAndPropagate(Node m, Node c)
		{
			if (m.Type != c.Type)
			{
				throw new ExecutionException("Different Labels");
			} else if (m.Type == Node.NodeType.Lam) {
				tg.AddUndirectEdge(m.Body, c.Body);
			} else if (m.Type == Node.NodeType.App)	{
				tg.AddUndirectEdge(m.Left, c.Left);
				tg.AddUndirectEdge(m.Right, c.Right);
			} else { // Node.NodeType.Var
				if (m.Binder == null && c.Binder == null) {}
				else if (m.Binder != null && c.Binder != null) {}
				else throw new ExecutionException("Not Homogeneous");
			}
			m.Canonic = c;
			c.Queue.Enqueue(m);
		}

		public static void LambdaCheck(TermGraph tg)
		{
			foreach (Node n in tg.Nodes)
			{
				Node c = n.Canonic;
				if (n.Type == Node.NodeType.Var && n != c)
				{
					if (n.Binder == null || c.Binder == null)
					{
						throw new ExecutionException("Not Open.");
					} else if (n.Binder.Canonic != c.Binder.Canonic)
					{
						throw new ExecutionException("Violation of Scopes.");
					}
				}
			}
		}

		public static void DAGCheck(TermGraph tg)
		{
			foreach (Node n in tg.Nodes)
				if (n.Canonic == null)
					BuildEquivalenceClass(n);
		}

		public static void Start(TermGraph Tg)
		{
			tg = Tg;
			tg.Print();
			DAGCheck(tg);
			LambdaCheck(tg);
		}
	}
}
